#!/usr/bin/env ruby
require 'set'
module Dhaka
  
  START_SYMBOL_NAME = "_Start_"
  END_SYMBOL_NAME = "_End_"
  
  class ProductionBuilder
    def initialize(grammar, symbol)
      @grammar = grammar
      @symbol = symbol
    end
  
    def method_missing(production_name, expansion, options = {})
      expansion_symbols = expansion.collect {|name| @grammar.symbols[name]}
      if precedence_symbol_name = options[:prec] 
        production = Production.new(@symbol, expansion_symbols, production_name.to_s, @grammar.symbol_for_name(precedence_symbol_name).precedence)
      else
        production = Production.new(@symbol, expansion_symbols, production_name.to_s)
      end
      @symbol.nullable = true if expansion_symbols.empty?
      @grammar.productions_by_symbol[production.symbol] << production
      raise "Duplicate production named #{production.name}" if @grammar.productions_by_name[production.name]
      @grammar.productions_by_name[production.name] = production
    end
  end

  class PrecedenceBuilder
    def initialize(grammar)
      @grammar = grammar
      @precedence_level = 0
    end
    def method_missing(associativity, symbol_names)
      symbol_names.each do |symbol_name|
        symbol = @grammar.symbols[symbol_name]
        symbol.precedence = Precedence.new(@precedence_level, associativity)
      end
      @precedence_level += 1
    end
  end

  class Grammar

    def self.inherited(grammar)
      class << grammar
        attr_accessor :symbols, :productions_by_symbol, :productions_by_name, :start_symbol, :end_symbol, :__first_cache
      end
      grammar.symbols = Hash.new {|hash, name| hash[name] = GrammarSymbol.new(name)}
      grammar.productions_by_symbol = Hash.new {|hash, name| hash[name] = Set.new([])}
      grammar.productions_by_name = {}
      grammar.end_symbol = grammar.symbols[END_SYMBOL_NAME]
      grammar.start_symbol = grammar.symbols[START_SYMBOL_NAME]
      grammar.__first_cache = {}
    end

    def self.for_symbol symbol, &blk
      symbol = symbols[symbol]
      symbol.non_terminal = true
      ProductionBuilder.new(self, symbol).instance_eval(&blk)
    end


    def self.productions
      productions_by_name.values
    end

    def self.production_named(name)
      productions_by_name[name]
    end

    def self.productions_for_symbol(symbol)
      productions_by_symbol[symbol]
    end

    def self.symbol_for_name(name)
      if symbols.has_key? name
        symbols[name]
      else
        raise "No symbol with name #{name} found"
      end
    end

    def self.terminal_symbols
      symbols.values.select {|symbol| symbol.terminal}
    end

    def self.non_terminal_symbols
      symbols.values.select {|symbol| symbol.non_terminal}
    end

    def self.closure(kernel)
      channels = Set.new

      result = compute_closure(kernel) do |hash, item|    
        if item.next_symbol and item.next_symbol.non_terminal
          productions_by_symbol[item.next_symbol].each do |production|
            channels << spontaneous_channel(item, hash[Item.new(production, 0)])
          end
        end
      end

      return channels, result
    end

    def self.precedences &blk
      PrecedenceBuilder.new(self).instance_eval(&blk)
    end

    def self.first(given_symbol)
      cached_result = self.__first_cache[given_symbol] 
      return cached_result if cached_result
      result = compute_closure([given_symbol]) do |hash, symbol|
            productions_by_symbol[symbol].each do |production| 
              symbol_index = 0
              while next_symbol = production.expansion[symbol_index]
                hash[next_symbol]
                break if !next_symbol.nullable
                symbol_index += 1
              end
            end if symbol.non_terminal
        end.values.select {|symbol| symbol.terminal}.to_set
      self.__first_cache[given_symbol] = result
      result
    end
  
    def self.spontaneous_channel(start_item, end_item)
      SpontaneousChannel.new(self, start_item, end_item)
    end

    def self.passive_channel(start_item, end_item)
      PassiveChannel.new(self, start_item, end_item)
    end

    def self.compute_closure(initial)
      closure_hash = ClosureHash.new do |hash, item|
        hash.dirty = true
        hash[item] = item
      end
      closure_hash.load_set(initial)
      while true
        closure_hash.keys.each do |element|
          yield closure_hash, element
        end
        break if !closure_hash.dirty
        closure_hash.dirty = false
      end
      return closure_hash
    end

  
  end
end