#!/usr/bin/env ruby
require 'set'
require 'logger'

module Dhaka
  class Parser
    include ParserMethods
    attr_reader :grammar, :start_state
  
    def initialize(grammar, logger = nil)
      if logger
        @logger = logger
      else
        @logger = Logger.new(STDOUT)
        @logger.level = Logger::WARN
      end
      @transitions = Hash.new {|hash, state| hash[state] = {}}
      @grammar = grammar
      @channels = []
      @states = Hash.new do |hash, kernel|
          channels, closure = @grammar.closure(kernel)
          @channels += channels.to_a
          new_state = ParserState.new(self, closure)
          hash[kernel] = new_state
          @logger.debug("Created #{new_state}.")
          new_state.transition_items.each do |symbol, items|
            destination_kernel = ItemSet.new(items.collect{|item| item.next_item})
            destination_state = hash[destination_kernel]
            items.each { |item| @channels << @grammar.passive_channel(item, destination_state.items[item.next_item]) }
            @transitions[new_state][symbol] = destination_state
          end
          new_state
      end
      initialize_states
    end

    def initialize_states
      start_productions = @grammar.productions_for_symbol(@grammar.start_symbol)
      raise NoStartProductionsError.new(@grammar) if start_productions.empty?
      start_items = ItemSet.new(start_productions.collect {|production| Item.new(production, 0)})
      start_items.each {|start_item| start_item.lookaheadset << @grammar.end_symbol}
      @start_state = @states[start_items]
      @logger.debug("Pumping #{@channels.size} channels...")
      pump_channels
      @logger.debug("Generating shift actions...")
      generate_shift_actions
      @logger.debug("Generating reduce actions...")
      generate_reduce_actions
    end

    def compile_to_ruby_source_as parser_class_name
      result = "class #{parser_class_name} < Dhaka::CompiledParser\n\n"
      result << "  self.grammar = #{@grammar.name}\n\n"
      result << "  start_with #{start_state.id}\n\n"
      states.each do |state|
        result << "#{state.compile_to_ruby_source}\n\n"
      end
      result << "end"
      result
    end

    def to_dot
      result = ["digraph x {", "node [fontsize=\"10\" shape=box size=\"5\"]"]
      result += states.collect { |state| state.to_dot }
      states.each { |state| 
          @transitions[state].each { |symbol, dest_state| 
              result << "#{state.dot_name} -> #{dest_state.dot_name} [label=\"#{symbol.name}\"]"
             }
         }
      result << ['}']
      result.join("\n")
    end
    def states
      @states.values
    end
  
    def generate_shift_actions
      @states.values.each do |state|
        @transitions[state].keys.each { |symbol|
            state.actions[symbol.name] = ShiftAction.new(@transitions[state][symbol])
           }
      end
    end

    def generate_reduce_actions
      @states.values.each do |state|  
        state.items.values.select{ |item| !item.next_symbol }.each do |item| 
          create_reduction_actions_for_item_and_state item, state
        end
      end
    end

    def create_reduction_actions_for_item_and_state item, state
      item.lookaheadset.each do |lookahead| 
        new_action = ReduceAction.new(item.production) 
        if existing_action = state.actions[lookahead.name]
          if ReduceAction === existing_action
            raise ParserReduceReduceConflictError.new(build_conflict_message(state, lookahead, new_action).join("\n"))
          else
            resolve_conflict state, lookahead, new_action
          end
        else
          state.actions[lookahead.name] = new_action
        end
      end
    end
  
  
    def resolve_conflict state, lookahead, new_action
      message = build_conflict_message(state, lookahead, new_action)
      shift_precedence = lookahead.precedence
      reduce_precedence = new_action.production.precedence
      if (shift_precedence && reduce_precedence)
        if (shift_precedence > reduce_precedence)
          message << "Resolving with precedence. Choosing shift over reduce."
        elsif (shift_precedence < reduce_precedence)
          message << "Resolving with precedence. Choosing reduce over shift."
          state.actions[lookahead.name] = new_action
        else
          case shift_precedence.associativity
            when :left
              message << "Resolving with left associativity. Choosing reduce over shift."
              state.actions[lookahead.name] = new_action
            when :right
              message << "Resolving with right associativity. Choosing shift over reduce."
            when :nonassoc
              message << "Resolving with non-associativity. Eliminating action."
              state.actions[lookahead.name] = nil
          end
        end
      else
        message << "No precedence rule. Choosing shift over reduce."
      end
      @logger.warn(message.join("\n"))
    end
  
    def build_conflict_message state, lookahead, new_action
      message = ["Parser Conflict at State:"] + state.items.values.collect{|it| it.to_s}
      message << "Existing: #{state.actions[lookahead.name]}"
      message << "New: #{new_action}"
      message << "Lookahead: #{lookahead}"
      message
    end
  
    def pump_channels
      while true
        unstable_count = 0
        @channels.each do |channel| 
          if channel.pump
            unstable_count += 1
          end 
        end
        break if unstable_count == 0
        @logger.debug("#{unstable_count} unstable channels...")
      end
    end
  
  end


  class ParserReduceReduceConflictError < StandardError
    def initialize(message)
      @message = message
    end
    def to_s
      @message
    end
  end
  
  class NoStartProductionsError < StandardError
    def initialize(grammar)
      @grammar = grammar
    end
    def to_s
      "No start productions defined for #{@grammar.name}"
    end
  end
end

