module Dhaka
  class UnrecognizedInputCharacterException < StandardError
    attr_reader :input, :char_index
    def initialize(input, char_index)
      @input = input
      @char_index = char_index
    end
    def to_s
      "Unrecognized character #{input[char_index].chr} encountered while tokenizing:\n #{input}"
    end
  end

  class TokenizerState
    attr_reader :actions

    def initialize
      @actions = {}
    end

    def for_characters(characters, &blk)
      characters.each do |character|
        actions[character] = blk
      end
    end

    def for_character(character, &blk)
      actions[character[0]] = blk
    end
  
    def to_s
      actions.inspect
    end

  end

  class Tokenizer

    def self.inherited(tokenizer)
      class << tokenizer
        attr_accessor :states
      end
      tokenizer.states = Hash.new {|hash, key| hash[key] = TokenizerState.new}
    end

    def self.for_state(state_name, &blk)
      states[state_name].instance_eval(&blk)
    end

    def self.tokenize(input)
      TokenizerRun.new(self, input).run
    end
  end

  class TokenizerRun

    attr_accessor :accumulator
    attr_reader :tokens
    def initialize(tokenizer, input)
      @tokenizer = tokenizer
      @input = input
      @current_state = tokenizer.states[:idle_state]
      @curr_char_index = 0
      @tokens = []
    end
  
    def run
      while curr_char
        blk = @current_state.actions[curr_char]
        raise UnrecognizedInputCharacterException.new(@input, @curr_char_index) unless blk
        instance_eval(&blk)
      end
      tokens
    end
  
    def curr_char
      @input[@curr_char_index] and @input[@curr_char_index].chr 
    end
  
    def advance
      @curr_char_index += 1
    end
  
    def switch_to state_name
      @current_state = @tokenizer.states[state_name]
    end

  end
end