#!/usr/bin/env ruby
require 'test/unit'
require 'simple_grammar'
require 'arithmetic_grammar'
require 'nullable_grammar'
require 'lalr_but_not_slr_grammar'
require 'rr_conflict_grammar'
require 'sr_conflict_grammar'

class ParserTest < Test::Unit::TestCase

  def setup
    @grammar = SimpleGrammar
    @parser = Dhaka::Parser.new(@grammar)
  end

  def test_parser_generates_states_with_correct_items
    expected_states = {}
    expected_states[1] = Set.new(['_Start_ ::= -> S # [_End_]',
                          'S ::= -> E [#]',
                          'E ::= -> E - T [#-]',
                          'E ::= -> T [#-]',
                          'T ::= -> n [#-]',
                          'T ::= -> ( E ) [#-]'])
    expected_states[2] = Set.new(['E ::= T -> [#)-]'])
    expected_states[3] = Set.new(['T ::= n -> [#)-]'])
    expected_states[4] = Set.new(['S ::= E -> [#]', 
                          'E ::= E -> - T [#-]'])
    expected_states[5] = Set.new(['_Start_ ::= S -> # [_End_]'])
    expected_states[6] = Set.new(['T ::= ( -> E ) [#)-]',
                          'E ::= -> E - T [)-]',
                          'E ::= -> T [)-]',
                          'T ::= -> n [)-]',
                          'T ::= -> ( E ) [)-]'])
    expected_states[7] = Set.new(['E ::= E - -> T [#)-]',
                          'T ::= -> n [#)-]',
                          'T ::= -> ( E ) [#)-]'])
    expected_states[8] = Set.new(['E ::= E - T -> [#)-]'])
    expected_states[9] = Set.new(['T ::= ( E -> ) [#)-]',
                          'E ::= E -> - T [)-]'])
    expected_states[10] = Set.new(['T ::= ( E ) -> [#)-]'])
    expected_states[11] = Set.new(['_Start_ ::= S # -> [_End_]'])
    actual_states = Set.new(@parser.states.collect {|state| Set.new(state.items.values.collect {|item| item.to_s})})
    #write_parser(@parser)
    expected_states.values.each do |state|
      assert set_finder(state, actual_states), "expected #{state.to_a}"
    end
  end
  def build_tokens(token_symbol_names, grammar)
    token_symbol_names.collect {|symbol_name| Dhaka::Token.new(grammar.symbol_for_name(symbol_name), nil)}
  end
  
  def test_parser_generates_syntax_tree_given_a_stream_of_symbols
    syntax_tree = @parser.parse(build_tokens(['(','n','-','(','n','-','n',')',')','-','n','#'], @grammar)).syntax_tree
    assert_equal \
      ["literal",
       "term",
       "literal",
       "term",
       "literal",
       "subtraction",
       "parenthetized_expression",
       "subtraction",
       "parenthetized_expression",
       "term",
       "literal",
       "subtraction",
       "expression",
       "start"], syntax_tree.linearize
  end
  
  def get_linearized_parse_result(input, parser)
    parser.parse(build_tokens(input, parser.grammar)).syntax_tree.linearize
  end
  
  def test_with_a_different_grammar_with_division
    grammar = ArithmeticGrammar
    parser = Dhaka::Parser.new(grammar)
    parser_input = ['(','n','-','(','n','/','n','-','n',')','/','n',')']
    assert_equal \
      ["getting_literals",
       "factor",
       "term",
       "getting_literals",
       "factor",
       "getting_literals",
       "division",
       "term",
       "getting_literals",
       "factor",
       "subtraction",
       "unpacking_parenthetized_expression",
       "factor",
       "getting_literals",
       "division",
       "subtraction",
       "unpacking_parenthetized_expression",
       "factor",
       "term",
       "expression"], get_linearized_parse_result(parser_input, parser)

    parser_input = ['h','(','(','n',')','-','n',',','n',')']
    assert_equal \
      ["max_function",
       "getting_literals",
       "factor",
       "term",
       "unpacking_parenthetized_expression",
       "factor",
       "term",
       "getting_literals",
       "factor",
       "subtraction",
       "getting_literals",
       "factor",
       "term",
       "single_args",
       "concatenating_args",
       "evaluating_function",
       "function",
       "factor",
       "term",
       "expression"], get_linearized_parse_result(parser_input, parser)
  end
  
  def test_with_a_grammar_with_nullables_after_terminals
    grammar = NullableGrammar
    parser = Dhaka::Parser.new(grammar)
    parser_input = ['(','a',')']
    assert_equal \
      ["literal_a",
       "empty_element_list",
       "concatenate_element_lists",
       "element_list",
       "tuple"], get_linearized_parse_result(parser_input, parser)
  end

  def test_with_a_grammar_that_is_not_SLR
    grammar = LALRButNotSLRGrammar
    parser = Dhaka::Parser.new(grammar)
    parser_input = ['b','d','c']
    assert_equal(["A_d", "E_bAc", "start"], get_linearized_parse_result(parser_input, parser))
  end

  def test_with_a_grammar_that_should_generate_an_RR_conflict
    grammar = RRConflictGrammar
    assert_raise(Dhaka::ParserReduceReduceConflictError) { Dhaka::Parser.new(grammar) }
  end

  def set_finder(set1, set2)
    set2.inject(false) {|result, member| result ||= member == set1}
  end
  
  def write_parser(parser)
    File.open('parser.dot', 'w') do |file|
      file << parser.to_dot
    end
  end
  
end
